/************************************************************************

 The contents of this file are subject to the Mozilla Public License
 Version 1.1 (the "License"); you may not use this file except in	
 compliance with the License. You may obtain a copy of the License at
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS"
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 License for the specific language governing rights and limitations
 under the License.

 The Original Code is Isearch.

 The Initial Developer of the Original Code is Clearinghouse for
 Networked Information Discovery and Retrieval.

 Portions created by Netscape Communications are Copyright (C) 2001. 
 All Rights Reserved.

 Portions created by the Clearinghouse for Networked Information
 Discovery and Retrieval are Copyright (C) 1994. All Rights Reserved.

************************************************************************/

/* A set of search results. */

#include <stdlib.h>

#include "defs.h"
#include "string.h"
#include "common.h"
#include "vlist.h"
#include "strlist.h"
#include "record.h"
#include "mdtrec.h"
#include "mdt.h"
#include "idbobj.h"
#include "iresult.h"
#include "irset.h"
#include "dtreg.h"
#include "catids.h"

extern MDTHASHTABLE ht;

#ifndef DOC_TYPE_KEYWORD 
#define DOC_TYPE_KEYWORD 99
#endif

// now we don't need any path-prefix
// const int pathprefixlen = 20;

int IrsetIndexCompare(const void* x, const void* y) 
{
  INT4 Difference = ( (*((PIRESULT)y)).MdtIndex - 
		     (*((PIRESULT)x)).MdtIndex );
  if (Difference < 0) {
    return (-1);
  } else {
    if (Difference == 0) {
      return(0);
    } else {
      return 1;
    }
  }
}

IRSET::IRSET(const PIDBOBJ DbParent) 
{
  Init(DbParent, 2000);
}

IRSET::IRSET(const PIDBOBJ DbParent, const int start_size) 
{
  int s = start_size;

  if (s > MAX_RESULT_ENTRIES) 
    s = MAX_RESULT_ENTRIES;
  else if (s < 100)
    s = 100;
  
  Init(DbParent, s);
}

void IRSET::Init(const PIDBOBJ DbParent, int start_size) 
{
  TotalEntries = 0;
  MaxEntries = start_size;
  Table = new IRESULT[MaxEntries];
  Parent = DbParent;
  MinScore=999999.0;
  MaxScore=0.0;
  INT ScoreSort=0;		// 1 if sorted by score

  MaxRecordLength = Parent->GetMaxRecordLength();
  TopLevelRestrictDir = Parent->GetTopLevelRestrictDir();
  RestrictDir = Parent->GetRestrictDir();
  RestrictAdult = Parent->GetRestrictAdult();
  HashInit();
}

void
IRSET::HashInit() {
  int i;

#ifdef DEBUG_HASH_COLLISIONS
  collisions = 0;
#endif

  // lets use a power of two for fast hashing
  // 2^15 = 32768
  tableSize = 32768;

  // allow for zero length table to disable hash table function
  if (tableSize > 0) {
    ht = (i_list **) malloc(tableSize * sizeof(i_list *));

    // zero the table
    memset((void *)ht, 0, sizeof(i_list *) *tableSize);
  }
}

int
IRSET::Hash(int mdt_index) {
  // thankfully the mdt_index is the filenumber and this means
  // we have a high level of dispirstion built in
  //
  // This is the worlds fasted hash funtion.
  // 0x7FFF = 2^15 - 1
  return (mdt_index & 0x7FFF);
}

int
IRSET::HashAdd(int mdt_index) {

  // XXX - probably want to just add this to the the regular Add() function

  int hash;
  i_list *node = NULL;

  // check to see if this str/offet pair already exists
  // in the hash table
  hash = Hash(mdt_index);
  node = ht[hash];
  while (node && (mdt_index != node->mdt_index)) 
    node = node->next;

  if (!node) {
    // we have a new entry, add it.

    // insert node into the hashtable in front
    node = (i_list *) malloc(sizeof(i_list));
    node->mdt_index = mdt_index;
    node->irset_index = TotalEntries;
    node->next = ht[hash];

#ifdef DEBUG_HASH_COLLISIONS
    if (ht[hash])
      collisions++;
#endif

    ht[hash] = node;

  }

  return (node->irset_index);

}

int
IRSET::HashGet(int mdt_index) {

  INT h = Hash(mdt_index);
  i_list *node = NULL;

  node = ht[h];
  while (node && (node->mdt_index != mdt_index))
      node = node->next;

  if (node)
    return (node->irset_index);
  else
    return -1;
}

void IRSET::SetRestrictDir(const STRING& DirName) {

  RestrictDir = DirName;
}

void IRSET::SetTopLevelRestrictDir(unsigned int TopLevelDirs) {

  TopLevelRestrictDir = TopLevelDirs;
}

void IRSET::SetRestrictAdult(int i) {

  RestrictAdult = i;
}

DOUBLE IRSET::GetMaxScore(){
  return(MaxScore);
}

DOUBLE IRSET::GetMinScore(){
  return(MinScore);
}

IRSET& IRSET::operator=(const IRSET& OtherIrset) {
  if (Table) {
    delete [] Table;
  }
  Init(OtherIrset.GetParent(),OtherIrset.GetTotalEntries());
  INT y = OtherIrset.GetTotalEntries();
  INT x;
  IRESULT iresult;

  for (x=1; x<=y; x++) {
    OtherIrset.GetEntry(x, &iresult);
    Table[x-1] = iresult;
  }
  return *this;
}

IRSET* IRSET::Duplicate() const {
  IRSET* Temp = new IRSET(Parent);
  *Temp = *(this);
  return Temp;
}


PIRSET IRSET::Duplicate(){
  IRSET * Temp= new IRSET(Parent);
  *Temp=*(this);
  return (Temp);
}


void IRSET::FastAddEntry(const IRESULT& ResultRecord, const INT AddHitCounts) 
{
  INT x;

  // fast add doesn't check for category match or adult searches
  // because this has already been done...

  x = HashGet(ResultRecord.MdtIndex);
  if (0 <= x) {
    // the entries already exists in the table
    // update the hitcounts
    if (AddHitCounts) {
      Table[x].IncScore(ResultRecord.GetScore());
      Table[x].IncHitCount(ResultRecord.HitCount);
      Table[x].IncRealHitCount(ResultRecord.RealHitCount);
    }
    Table[x].IncScore(ResultRecord.GetScore());

    return;
  }

  // otherwise its a new record, so add it
  if (TotalEntries == MaxEntries)
    Expand();
  if (TotalEntries < MaxEntries) {
    x = HashAdd(ResultRecord.MdtIndex);
    Table[TotalEntries] = ResultRecord;
    TotalEntries = TotalEntries + 1;
  }
#if 0
  else {
      cerr << "FULL: Dropping Entry = " << x << " "
	   << " MDT index= " << ResultRecord.MdtIndex << " "
	   << " Path number= " << ResultRecord.pathName << " "
	   << " File number= " << ResultRecord.fileName << "\n";
  }
#endif  
  
}

void IRSET::AddEntry(const IRESULT& ResultRecord, const INT AddHitCounts) 
{
  INT x;
  MDTREC mdtrec;

  // No need to check the Category or Adult status, because
  // if its in the hash its already cleared the test

  x = HashGet(ResultRecord.MdtIndex);

  if ((x >= 0) && (x < TotalEntries)) {
    // the entries already exists in the table
    // update the hitcounts
    if (AddHitCounts) {
      Table[x].IncHitCount(ResultRecord.HitCount);
    }
    Table[x].IncScore(ResultRecord.GetScore());

#if 0
    cerr << "DUP Entry = " << x << " "
	 << " MDT index= " << Table[x].MdtIndex << " "
	 << " Path number= " << Table[x].pathName << " "
	 << " File number= " << Table[x].fileName << "\n";
#endif

    return;
  }

  // otherwise its a new record, so add it

  // First test to see if the new entry has the same top level
  // category as the restricted Directory
  if (TopLevelRestrictDir != TOP_CAT_None) 
     if (TopLevelRestrictDir != ResultRecord.TopLevelCat)
       return;

  // ignore all records that are too long
  if (MaxRecordLength && 
      ((ResultRecord.GlobalFileEnd - ResultRecord.GlobalFileStart) > 
       MaxRecordLength))
    return;

  // first we fetch the Path of the file so we can check
  // what category we are dealing with
  STRING s;
  if (RestrictDir != "") {
    // only look up the path if we need it
    Parent->GetMainMdt()->GetEntry(ResultRecord.MdtIndex, &mdtrec);
    mdtrec.GetPathName(&s);
  }

  //
  // Only add Adult links if we are doing Adult searches
  //
  if ((RestrictAdult == 0) || (ResultRecord.TopLevelCat != TOP_CAT_Adult)) {

    //
    // Only add the record if it falls in the designated category
    //
    //

    //
    // I have removed the need to prefix the "filename" of each site and
    // category with the path name. It wasn't used anymore since the
    // BFB was introduced. I finally got arround to fixing this.
    //
    // Also the ODP need the base category to match exactly, thus
    // the (s.Search(RestrictDir) == 0). If you want a  ( XXX >= 0)
    // we should figure out a way to parameterize this.
    //
    if ((RestrictDir == "") || (s.Search(RestrictDir) == 1)) {
      if (TotalEntries == MaxEntries)
	Expand();
      if (TotalEntries < MaxEntries) {
	// only add the entry if we aren't past our max IRSET size

	// add the entry to the hash table
	x = HashAdd(ResultRecord.MdtIndex);

	// add the entry to the table
	Table[TotalEntries] = ResultRecord;

#if 0
	cerr << "Entry = " << TotalEntries << " "
	     << " MDT index= " << Table[TotalEntries].MdtIndex << " "
	     << " Path number= " << Table[TotalEntries].pathName << " "
	     << " File number= " << Table[TotalEntries].fileName << "\n";
#endif  

	TotalEntries++;

      }
    }
  }
}

void IRSET::GetEntry(const INT Index, PIRESULT ResultRecord) const {
  if ((Index > 0) && (Index <= TotalEntries) ) {
    *ResultRecord = Table[Index-1];
  }
}

void IRSET::Expand() {
  if (TotalEntries < MAX_RESULT_ENTRIES) 
    Resize(MAX_RESULT_ENTRIES);
}

void IRSET::CleanUp() {
  Resize(TotalEntries);
}

void IRSET::Resize(const INT Entries) {

  // only resize if we have to
  if (Entries <= MaxEntries)
    return;

  PIRESULT Temp = new IRESULT[Entries];
  INT RecsToCopy;
  INT x;
  if (Entries >= TotalEntries) {
    RecsToCopy = TotalEntries;
  } else {
    RecsToCopy = Entries;
    TotalEntries = Entries;
  }
  for (x=0; x<RecsToCopy; x++) {
    // Not sure if Temp[x] = Table[x] is good enough.
    Temp[x] = Table[x];
  }
  if (Table)
    delete [] Table;
  Table = Temp;
  MaxEntries = Entries;
}

INT IRSET::GetTotalEntries() const {
  return TotalEntries;
}

INT IRSET::GetHitTotal() const {
  INT x;
  INT Total = 0;
  for (x=0; x<TotalEntries; x++) {
    Total += Table[x].HitCount;
  }
  return Total;
}

void IRSET::Or(IRSET * OtherIrset) {
  INT x;
  INT t = OtherIrset->GetTotalEntries();
  IRESULT OtherIresult;
  for (x=1; x<= t; x++) {
    OtherIrset->GetEntry(x, &OtherIresult);
    FastAddEntry(OtherIresult, 1);

#if 0
      cerr << "OR Entry = " << x << " "
	   << " MDT index= " << OtherIresult.MdtIndex << " "
	   << " Path number= " << OtherIresult.pathName << " "
	   << " File number= " << OtherIresult.fileName << "\n";
#endif  


  }
}

// A faster version by Bryn Dole, needs something extra to actually work though.. see comment.
void IRSET::CharProx(const IRSET& OtherIrset, const INT Distance) {

  IRESULT OtherIresult;
  IRSET MyResult(Parent);
  INT x = 1;
  INT y;
  INT count=0;
  INT t = OtherIrset.GetTotalEntries();
  
  while (x <= t) {
    OtherIrset.GetEntry(x, &OtherIresult);

    y = HashGet(OtherIresult.MdtIndex);
    if (y >= 0) {
      // terms are in the same record

      // XXX - Need to add back a list of hits per record with the possition of
      // XXX - the terms so that we can see if the terms are within a cetain
      // XXX - distance of each other

      //if ( (abs(OtherIresult.GlobalFileStart - Table[y].GlobalFileEnd) <= Distance) ||
      //(abs(OtherIresult.GlobalFileEnd - Table[y].GlobalFileStart) <= Distance)) {

      // terms are withing X characters of each other
      Table[y].IncHitCount(OtherIresult.HitCount);
      Table[y].IncRealHitCount(OtherIresult.RealHitCount);
      Table[y].IncScore(OtherIresult.GetScore());
      MyResult.FastAddEntry(Table[y], 0);
      count++;
    }
    x++;
  }

  delete [] Table;
  TotalEntries= count;
  MaxEntries= MyResult.MaxEntries;
  Table = MyResult.Table;
  tableSize = MyResult.tableSize;
  MyResult.Table = NULL;
  ht = MyResult.ht;
}

// faster AndNot() by Bryn Dole
void IRSET::AndNot(IRSET *OtherIrset) 
{
  IRESULT Iresult;
  IRSET MyResult(Parent);
  INT x = 1;
  INT count=0;
  INT t = GetTotalEntries();
  
  while (x <= t) {
    GetEntry(x, &Iresult);

    if (OtherIrset->HashGet(Iresult.MdtIndex) < 0) {
      // the entry was not found, so include it
      MyResult.FastAddEntry(Iresult, 0);
      count++;
    }
    x++;
  }

  delete [] Table;
  TotalEntries=count;
  MaxEntries=MyResult.MaxEntries;
  Table = MyResult.Table;
  tableSize = MyResult.tableSize;
  MyResult.Table = NULL;
  ht = MyResult.ht;
}

// Faster And() with merging implemented by Bryn Dole
void IRSET::And(IRSET * OtherIrset) 
{
  IRESULT OtherIresult;
  IRSET MyResult(Parent);
  INT x = 1;
  INT y;
  INT count=0;
  INT t = OtherIrset->GetTotalEntries();
  
  while (x <= t) {
    OtherIrset->GetEntry(x, &OtherIresult);

    y = HashGet(OtherIresult.MdtIndex);
    if (y >= 0) {
      // the entry was found, so include it
      Table[y].IncHitCount(OtherIresult.HitCount);
      Table[y].IncRealHitCount(OtherIresult.RealHitCount);
      Table[y].IncScore(OtherIresult.GetScore());
      MyResult.FastAddEntry(Table[y], 0);

#if 0
      cerr << "AND Entry = " << count << " "
	   << " MDT index= " << Table[y].MdtIndex << " "
	   << " Path number= " << Table[y].pathName << " "
	   << " File number= " << Table[y].fileName << "\n";
#endif  

      count++;
    }
    x++;
  }

  delete [] Table;
  TotalEntries= count;
  MaxEntries= MyResult.MaxEntries;
  Table = MyResult.Table;
  tableSize = MyResult.tableSize;
  MyResult.Table = NULL;
  // free(ht); and all associated buckets
  ht = MyResult.ht;
}

void IRSET::ComputeScores(const DOUBLE TermWeight, int total_count) {
  if (TotalEntries == 0) {
    return;
  }

  DTREG    dtreg(0);
  PDOCTYPE doctype;

  DOUBLE Weight;
  DOUBLE DocsInRs = TotalEntries;
  DOUBLE DocsInDb = Parent->GetMainMdt()->GetTotalEntries();
  DOUBLE InvDocFreq = DocsInDb / DocsInRs;
  DOUBLE SumSqScores = 0;
  DOUBLE SqrtSum;
  DOUBLE Score;
  int    x;

  doctype = dtreg.GetDocTypePtr(Table[1].DocumentType);

  //
  // Set the Term's weight of the term inversely proportional to the 
  // number of hits
  //
  // Weight = [1/MAX_RESULT_ENTRIES, 1 ]  = (0,1]
  if (total_count) {
      if (total_count > MAX_RESULT_ENTRIES)
	  total_count = MAX_RESULT_ENTRIES;
      Weight = ( 1.0 + (MAX_RESULT_ENTRIES - total_count)) / (DOUBLE)MAX_RESULT_ENTRIES;
  }
  else {
      Weight = ( 1.0 + (MAX_RESULT_ENTRIES - TotalEntries)) / (DOUBLE)MAX_RESULT_ENTRIES;
  }

  for (x=0; x<TotalEntries; x++) {
      // IncDocFreq measures the inverse frequency of the term in the entire DB
      Score = doctype->ScoreHit(Table[x] , InvDocFreq);
      Table[x].SetScore(Score);
      SumSqScores += (Score * Score);
  }

  SqrtSum = sqrt(SumSqScores);
  if (SqrtSum == 0.0) {
      SqrtSum = 1.0;
  }

  for (x=0; x<TotalEntries; x++) {
      // TermWeight & Weight apply to all members of a set
      Score= ( Table[x].GetScore() / SqrtSum ) * Weight * TermWeight;
      if(Score>MaxScore)
	  MaxScore=Score;
      if(Score<MinScore)
          MinScore=Score;
      Table[x].SetScore(Score);
  }
}

int IrsetScoreCompare(const void* x, const void* y) {

  DOUBLE Difference;
  int  rankDifference = ((*((PIRESULT)y)).RealHitCount - 
			 (*((PIRESULT)x)).RealHitCount );

  if (rankDifference < 0) {
    return (-1);
  } 
  else {
    if ( rankDifference > 0) {
      return (1);
    }
    else {
      // Use the regular score to break the tie
      
      Difference = ((*((PIRESULT)y)).GetScore() - 
		    (*((PIRESULT)x)).GetScore() );
      
      if (Difference < 0) {
	return (-1);
      } else {
	if (Difference == 0) {
	  // use the record number to break the tie
	  return ((*((PIRESULT)x)).MdtIndex - 
		  (*((PIRESULT)y)).MdtIndex );
	} else {
	  return 1;
	}
      }
    }
  }
}

void IRSET::SortByScore() {

  DOUBLE Score;
  DOUBLE NewMaxScore = MaxScore;
  int x;

  // XXX - new for newhoo, give presidence to matches with
  //       the highest number of term hits
  for (x=0; x<TotalEntries; x++) {

    Score =  Table[x].Score + Table[x].RealHitCount * MaxScore;

    if( Score > NewMaxScore)
      NewMaxScore=Score;

    Table[x].SetScore(Score);

  }

  // adjust maximum score
  MaxScore=NewMaxScore;

  qsort(Table, TotalEntries, sizeof(IRESULT), IrsetScoreCompare);


#if 0 
  // Score debugging
  for (x=0; (x < TotalEntries) && ( x < 1000); x++) {
    cerr << Table[x].RealHitCount << "  " 
	 << Table[x].HitCount  << "  " 
	 << Table[x].Score << "  "
	 << "\n";
  }
#endif

  ScoreSort=1;
}

void IRSET::SortByIndex() {
  qsort(Table, TotalEntries, sizeof(IRESULT), IrsetIndexCompare);
  ScoreSort=0;
}

void IRSET::SetParent(PIDBOBJ const NewParent) {
  Parent = NewParent;
}

PIDBOBJ IRSET::GetParent() const {
  return Parent;
}

INT IRSET::GetScaledScore(const DOUBLE UnscaledScore, const INT ScaleFactor) {
	DOUBLE Diff = MaxScore - MinScore;
	if (Diff == 0) {
	  // no variance in scores
	  Diff=1;
	  // return max score
	  return (ScaleFactor);
	}
	return ( (INT)( ((UnscaledScore - MinScore) * ScaleFactor) / Diff ) );
}

IRSET::~IRSET() {

#ifdef DEBUG_HASH_COLLISIONS
  if (TotalEntries)
    cerr << "Total collisions = " << collisions << "   " 
	 << TotalEntries << "  " << (( 100 * collisions) / TotalEntries)
	 << "\n";
#endif

  if (Table)
    delete [] Table;
}
